"""
Module containing generic utils functions.
"""
import hashlib
import ssl
from pathlib import Path
from shutil import unpack_archive
from typing import Callable, Iterable, Union
from urllib import request

import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import torch
from matplotlib import cm
from matplotlib.colors import Normalize  # type: ignore
from PIL import Image  # type: ignore
from sklearn.manifold import TSNE  # type: ignore
from tqdm import tqdm  # type: ignore


def download_file(url: str, path: Path, print_progress: bool) -> None:
    """Download file from URL to path.

    Args:
        url: the url from which to download the file.
        path: the path and file name for the downloaded file.
        print_progress: if True print a progress bar while downloading the file.
    """

    def updater_callback() -> Callable[[int, int, int], None]:
        pbar = tqdm(total=None)

        def bar_update(count: int, block_size: int, total_size: int) -> None:
            if pbar.total is None and total_size:
                pbar.total = total_size
            progress_bytes = count * block_size
            pbar.update(progress_bytes - pbar.n)

        return bar_update

    ssl._create_default_https_context = ssl._create_unverified_context
    fn_callback = updater_callback() if print_progress else None
    request.urlretrieve(url, path, reporthook=fn_callback)


def check_integrity(path: Path, md5_hash: str = str(None)) -> bool:
    """Check the integrity of a file.

    The integrity of the file is ensured by checking if the path points to a file. Eventually,
    the md5 digest will be checked if it is not None.

    Args:
        path: the path to the file.
        md5_hash: the Defaults to None.

    Returns:
        True if the check is passed.
    """
    if not path.is_file:
        return False

    if md5_hash is not None:
        return md5_hash == get_md5(path)

    return True


def get_md5(path: Path, chunk_size: int = 1024 * 1024) -> str:
    """Compute the MD5 hash value, to handle big files the function splits the file in chunks.

    Args:
        path: the path to the file.
        chunk_size: the size for the chunk (multiple of 128). Defaults to 1024*1024.

    Returns:
        The hex string representation for the MD5 digest.
    """
    hash_md5 = hashlib.md5()

    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()


def download_and_extract(
    url: str,
    path_file_downloaded: Path,
    path_destination: Path,
    md5_hash: str = str(None),
) -> None:
    """Download a compressed file from the url and extract it into path_destination.

    Args:
        url: the url from which download the file.
        path_file_downloaded: the path in which to save the downloaded file.
        path_destination: the path in which extract the content of the file.
        md5_hash (optional): the hash digest to use to check the integrity of the file.

    Raises:
        RuntimeError: if the downloaded file is not correct.
    """
    path_file_downloaded.parent.mkdir(exist_ok=True)
    download_file(url, path_file_downloaded, True)

    if not check_integrity(path_file_downloaded, md5_hash):
        raise RuntimeError(f"File {str(path_file_downloaded)} not found or corrupted.")

    unpack_archive(str(path_file_downloaded), path_destination)
    path_file_downloaded.unlink()


def get_tsne(
    data: Union[torch.Tensor, np.ndarray],
    n_components: int = 1,
    verbose: int = 1,
    perplexity: float = 40.0,
    n_iter: int = 300,
    random_state: int = 0,
) -> torch.Tensor:
    """Compute the t-SNE for each sample. NaN values converted to zero.

    Args:
        data: The input data with shape (N, N_DIMS)
        n_components: The component for the embedded space. Defaults to 1.
        verbose: The verbosity. Defaults to 1.
        perplexity: The perplexity is related to the number of nearest neighbors used.
                    Defaults to 40.0.
        n_iter: Maximum iterations for the optimization. Should be at least 250. Defaults to 300.
        random_state: The random number generator. Defaults to 0.

    Returns:
        The computed t-SNE with shape (N, n_components) normalized.
    """
    if isinstance(data, torch.Tensor):
        data = data.cpu().numpy()

    data = np.nan_to_num(data)  # type: ignore
    tsne = TSNE(
        n_components=n_components,
        verbose=verbose,
        perplexity=perplexity,
        n_iter=n_iter,
        random_state=random_state,
    )
    results = np.squeeze(tsne.fit_transform(data))
    min_val = np.min(results)  # type: ignore
    max_val = np.max(results)  # type: ignore
    normalized_results = (results - min_val) / (max_val - min_val)

    return torch.tensor(normalized_results)


def apply_color_map(
    values: Union[torch.Tensor, np.ndarray],
    name_map: str = "Spectral",
) -> torch.Tensor:
    """Map each scalar value in values to a color using a color map in matplotlib.

    Args:
        values: The input values as a torch tensor with shape (N,).
        name_map: The color map to use. Defaults to "Spectral".

    Returns:
        RGB color for each sample with shape (N, 3).
    """
    if isinstance(values, torch.Tensor):
        values = values.cpu().numpy()
    cmap = cm.get_cmap(name_map)
    colored = cmap(values)[:, :3]
    return torch.Tensor(colored)


def get_conf_matrix_img(conf_matrix: np.ndarray, size: int = 600) -> Image.Image:
    """Create a confusion matrix image.

    Args:
        conf_matrix: The confusion matrix.
        size: The output image size in pixels. Defaults to 600.

    Returns:
        The confusion matrix image as PIL Image.
    """
    c_mat = np.nan_to_num(conf_matrix).astype(np.float32)  # type: ignore
    h, w = c_mat.shape[0], c_mat.shape[1]

    for i in range(h):
        vmin = np.min(c_mat[i, :])  # type: ignore
        vmax = np.max(c_mat[i, :])  # type: ignore
        if vmax > 0:
            c_mat[i, :] = (c_mat[i, :] - vmin) / (vmax - vmin)

    zoom = size // h

    img = np.empty((h * zoom, w * zoom), dtype=np.float32)

    for i in range(h):
        for j in range(w):
            hstart = i * zoom
            hend = (i + 1) * zoom
            wstart = j * zoom
            wend = (j + 1) * zoom
            img[hstart:hend, wstart:wend] = c_mat[i, j]

    normalize = Normalize()
    cmap = cm.get_cmap("Blues")
    img = cmap(normalize(img))

    dpi = 100
    figsize = ((w * zoom) / dpi, (h * zoom) / dpi)
    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    fontsize = max(zoom / 10, 8)

    ax.set_xticks(np.arange(0, w) * zoom + (zoom / 2))
    ax.set_xticklabels([str(i) for i in range(w)], fontsize=fontsize)
    ax.set_yticks(np.arange(0, h) * zoom + (zoom / 2))
    ax.set_yticklabels([str(i) for i in range(h)], fontsize=fontsize)

    ax.imshow(img)

    for i in range(h):
        ax.plot(
            [0, w * zoom - 1],
            [(i * zoom + (zoom / 2)), (i * zoom + (zoom / 2))],
            linestyle=":",
            c="black",
            alpha=0.1,
        )

    for i in range(w):
        ax.plot(
            [(i * zoom + (zoom / 2)), (i * zoom + (zoom / 2))],
            [0, h * zoom - 1],
            linestyle=":",
            c="black",
            alpha=0.1,
        )

    fig.tight_layout()
    fig.canvas.draw()
    pilimg = Image.frombytes("RGB", fig.canvas.get_width_height(), fig.canvas.tostring_rgb())

    plt.close(fig)

    return pilimg


def progress_bar(iterable: Iterable, desc: str = "", num_cols: int = 60) -> Iterable:
    """Decorate an iterable object using a progress bar.

    Args:
        iterable: the iterable to decorate.
        desc: the description to print. Defaults to "".
        num_cols: The width of the entire output message. Defaults to 60.

    Returns:
        The decorated iterable.
    """
    bar_format = "{percentage:.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]"
    if len(desc) > 0:
        bar_format = "{desc}: " + bar_format
    return tqdm(iterable, desc=desc, bar_format=bar_format, ncols=num_cols, leave=False)



from typing import Dict, List, Tuple

import torch

from pycarus.metrics.metric import Metric


class PartSegmentationIoU(Metric):
    def __init__(
        self,
        use_only_category_logits: bool,
        category_to_parts_map: Dict[str, List[int]],
    ) -> None:
        """Create a Metric to compute part segmentation IoU.

        Args:
            use_only_category_logits: Whether to compute predictions considering only
                                      the logits corresponding to the parts belonging
                                      to the current category.
            category_to_parts_map: A dictionary that maps classes to parts labels.
                                   E.G. {"airplane": [0, 1, 2, ...], ...}
        """
        self.use_only_category_logits = use_only_category_logits

        self.category_to_parts_map = category_to_parts_map

        self.part_to_category_map: Dict[int, str] = {}
        for cat in category_to_parts_map:
            for part in category_to_parts_map[cat]:
                self.part_to_category_map[part] = cat

        self.ious: Dict[str, List[float]] = {}

        super().__init__()

    def reset(self) -> None:
        """Reset metric."""
        self.ious = {cat: [] for cat in self.category_to_parts_map}

    def update(self, predictions: torch.Tensor, targets: torch.Tensor) -> None:
        """Update metric.

        Args:
            predictions: A batch of predictions with shape (B, NUM_POINTS, NUM_CLASSES).
            targets: A batch of targets with shape (B, NUM_POINTS).
        """
        batch_size = predictions.shape[0]

        for b in range(batch_size):
            logits = predictions[b]
            cat = self.part_to_category_map[int(targets[b, 0])]
            parts = self.category_to_parts_map[cat]

            if self.use_only_category_logits:
                logits = logits[:, parts]
                pred_labels_per_point = logits.max(1)[1] + parts[0]
            else:
                pred_labels_per_point = logits.max(1)[1]

            part_ious: List[float] = [0.0 for _ in range(len(parts))]
            for p in parts:
                union = torch.sum((targets[b] == p) | (pred_labels_per_point == p))
                if union == 0:
                    # part is not present in this shape
                    part_ious[p - parts[0]] = 1.0
                else:
                    intersection = torch.sum((targets[b] == p) & (pred_labels_per_point == p))
                    part_ious[p - parts[0]] = float(intersection) / (float(union) + 1e-10)

            self.ious[cat].append(sum(part_ious) / (len(part_ious)) + 1e-10)

    def compute(self) -> Tuple[float, float]:
        """Compute final results.

        Returns:
            - Mean Class IoU: the average mIoU for each class.
            - Mean Instance IoU: the average IoU for each sample.
        """
        instance_ious: List[float] = []
        mIoUs: Dict[str, float] = {}

        for cat in self.ious:
            for iou in self.ious[cat]:
                instance_ious.append(iou)
            mIoUs[cat] = sum(self.ious[cat]) / (len(self.ious[cat]) + 1e-10)

        mIoU_per_cat = list(mIoUs.values())
        mean_class_iou = sum(mIoU_per_cat) / (len(mIoU_per_cat) + 1e-10)

        mean_instance_iou = sum(instance_ious) / (len(instance_ious) + 1e-10)

        return mean_class_iou, mean_instance_iou



from pathlib import Path
from typing import Callable, List, Tuple

import h5py  # type: ignore
import torch
from torch import Tensor
from torch.utils.data import Dataset

from pycarus.transforms.var import Compose


class ShapeNetCoreV2(Dataset):
    def __init__(
        self,
        root: Path,
        split: str,
        download: bool = False,
        transforms: List[Callable] = [],
    ) -> None:
        """Create ShapeNetCoreV2 dataset.

        Args:
            root: The root that contains the directory "shapenetcorev2".
            split: The required split, it can be "train", "val", "test" or a combination of them,
                   i.e. "trainval", "traintest" or "trainvaltest".
            download: If True download the dataset using the "tmp" folder. Defaults to False.
            transforms: The list of transforms to apply to pcds. Defaults to [].
        """
        super().__init__()

        assert split in ["train", "test", "val", "trainval", "traintest", "trainvaltest"]
        self.split = split

        self.dataset_root = root / "shapenetcorev2"

        if download:
            self.download()

        h5_files: List[Path] = []
        for s in ["train", "val", "test"]:
            if s in self.split:
                h5_files.extend(list(self.dataset_root.glob(f"*{s}*.h5")))
        h5_files.sort()

        self.pcds, self.labels = self.load_h5(h5_files)

        self.transform = Compose(transforms)

        with open(self.dataset_root / "shape_names.txt") as f:
            self.shape_names = [line.strip() for line in f.readlines()]

    @staticmethod
    def load_h5(paths: List[Path]) -> Tuple[Tensor, Tensor]:
        """Load all the given h5 files.

        Args:
            paths: The list of paths to the h5 files to load.

        Returns:
            - One tensor with all the pcds, with shape (NUM_PCDS, NUM_POINTS, 3).
            - One tensor with all the labels, with shape (NUM_PCDS,).
        """
        all_pcds: List[Tensor] = []
        all_labels: List[Tensor] = []
        for h5_path in paths:
            f = h5py.File(h5_path, "r+")
            pcds = f["data"][:].astype("float32")
            labels = f["label"][:].astype("int64")
            f.close()
            all_pcds.append(torch.from_numpy(pcds))
            all_labels.append(torch.from_numpy(labels))
        return torch.cat(all_pcds, dim=0), torch.cat(all_labels, dim=0).squeeze(1)

    def __len__(self) -> int:
        """Return the number of pcds in the dataset.

        Returns:
            The number of pcds in the dataset.
        """
        return self.pcds.shape[0]

    def __getitem__(self, index: int) -> Tuple[Tensor, Tensor]:
        """Return the pcd and the label at the given index.

        Args:
            index: The index of the required element.

        Returns:
            - A tensor with the point cloud with shape (NUM_POINTS, 3).
            - A tensor with the label.
        """
        pcd, label = self.pcds[index], self.labels[index]

        pcd = self.transform(pcd)

        return pcd, label

    def download(self) -> None:
        """Function to download the dataset."""
        raise NotImplementedError("Function not implemented.")



import torch
from torch import nn
from torch.autograd import Function

"""
Chamfer Distance for 3D points CUDA extension.

Code from:
https://github.com/ThibaultGROUEIX/ChamferDistancePytorch

Author: Thibault Groueix
"""

import chamfer_cuda  # type: ignore


class Chamfer_GPUFunction(Function):
    @staticmethod
    def forward(ctx, xyz1, xyz2):  # type: ignore
        batchsize, n, dim = xyz1.size()
        assert (
            dim == 3
        ), "Wrong last dimension for the chamfer distance 's input! Check with .size()"
        _, m, dim = xyz2.size()
        assert (
            dim == 3
        ), "Wrong last dimension for the chamfer distance 's input! Check with .size()"
        device = xyz1.device

        device = xyz1.device

        dist1 = torch.zeros(batchsize, n)
        dist2 = torch.zeros(batchsize, m)

        idx1 = torch.zeros(batchsize, n).type(torch.IntTensor)  # type: ignore
        idx2 = torch.zeros(batchsize, m).type(torch.IntTensor)  # type: ignore

        dist1 = dist1.to(device)
        dist2 = dist2.to(device)
        idx1 = idx1.to(device)
        idx2 = idx2.to(device)
        torch.cuda.set_device(device)

        chamfer_cuda.forward(xyz1, xyz2, dist1, dist2, idx1, idx2)
        ctx.save_for_backward(xyz1, xyz2, idx1, idx2)
        return dist1, dist2, idx1, idx2

    @staticmethod
    def backward(ctx, graddist1, graddist2, gradidx1, gradidx2):  # type: ignore
        xyz1, xyz2, idx1, idx2 = ctx.saved_tensors
        graddist1 = graddist1.contiguous()
        graddist2 = graddist2.contiguous()
        device = graddist1.device

        gradxyz1 = torch.zeros(xyz1.size())
        gradxyz2 = torch.zeros(xyz2.size())

        gradxyz1 = gradxyz1.to(device)
        gradxyz2 = gradxyz2.to(device)
        chamfer_cuda.backward(xyz1, xyz2, gradxyz1, gradxyz2, graddist1, graddist2, idx1, idx2)
        return gradxyz1, gradxyz2


class Chamfer_GPU(nn.Module):
    def __init__(self) -> None:
        super(Chamfer_GPU, self).__init__()  # type: ignore

    def forward(self, input1, input2):  # type: ignore
        input1 = input1.contiguous()
        input2 = input2.contiguous()
        return Chamfer_GPUFunction.apply(input1, input2)
